"""Add IPs to blocklist."""
import datetime
import logging

from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand
from django.core.validators import validate_ipv46_address

from ...models import BlockedIP
from ...utils import COOLDOWN

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("ips", nargs="+", type=str, help="IPs (space-separated) to add")
        parser.add_argument(
            "--cooldown",
            help=f"Days with no connections before IP is dropped from blocklist (default: {COOLDOWN})",
        )
        parser.add_argument("--reason", help="'reason' field value for the added IPs", default="")

    help = __doc__

    def handle(self, *args, **options):
        ips = options.get("ips")
        cooldown = options.get("cooldown")
        reason = options.get("reason")
        now = datetime.datetime.now()
        for ip in ips:
            try:
                validate_ipv46_address(ip)
            except ValidationError:
                print(f"Invalid IP: {ip}")
                continue
            entry, created = BlockedIP.objects.get_or_create(ip=ip)
            if created:
                entry.reason = reason
                entry.cooldown = cooldown or COOLDOWN
                entry.first_seen = now
                entry.save()
                print(f"Created entry for {ip}")
            else:
                print(f"{ip} already listed; use update_blocklist to update")



"""Update specified IPs with new last-seen, reason, or cooldown."""
import logging

from django.core.management.base import BaseCommand

from ...models import BlockedIP

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("ips", nargs="+", type=str, help="IPs (space-separated) to update")
        parser.add_argument("--cooldown")
        parser.add_argument("--reason")
        parser.add_argument("--last-seen", help="Datetime in ISO8601 format")

    help = __doc__

    def handle(self, *args, **options):
        ips = options.get("ips")
        cooldown = options.get("cooldown")
        reason = options.get("reason")
        last_seen = options.get("last_seen")
        for ip in ips:
            try:
                entry = BlockedIP.objects.get(ip=ip)
            except BlockedIP.DoesNotExist:
                print(f"{ip} not found")
                continue
            else:
                if reason:
                    entry.reason = reason
                if cooldown:
                    entry.cooldown = cooldown
                if last_seen:
                    entry.last_seen = last_seen
                entry.save()
                print(f"Updated existing entry for {ip}")



from unittest.mock import patch

from django.core.cache import cache
from django.test import TestCase, override_settings

from ..models import BlockedIP
from ..utils import CACHE_KEY, add_to_blocklist


class SettingsTests(TestCase):
    def tearDown(self) -> None:
        BlockedIP.objects.all().delete()
        cache.delete(CACHE_KEY)

    # override_settings doesn't work for this one, so we patch instead
    @patch("django_blocklist.utils.COOLDOWN", 11)
    def test_cooldown_setting(self):
        add_to_blocklist(["1.1.1.1"])
        blocked = BlockedIP.objects.get(ip="1.1.1.1")
        self.assertEqual(blocked.cooldown, 11)

    @override_settings(BLOCKLIST_CONFIG={"denial-template": "418"})
    def test_message_setting(self):
        add_to_blocklist(["127.0.0.1"])
        response = self.client.get("/")
        self.assertEqual(response.status_code, 400)
        self.assertIn(response.content, b"418")



"""Look up provided IPs in the blocklist. Set --verbosity=2 to see full record for found IPs."""
import logging
import sys

from django.conf import settings
from django.core.management.base import BaseCommand

from ...models import DEFAULT_COOLDOWN, BlockedIP

logger = logging.getLogger(__name__)
DEFAULT_DAYS = settings.BLOCKLIST_CONFIG.get("cooldown") or DEFAULT_COOLDOWN


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument("ips", nargs="+", type=str, help="IPs to look up (space separated)")

    def handle(self, *args, **options):
        ips = options.get("ips")
        self.verbosity = options.get("verbosity", 1)
        for entry in BlockedIP.objects.filter(ip__in=ips):
            print(entry.verbose_str() if self.verbosity >= 2 else f"Found: {entry.ip}")
        if self.verbosity > 1:
            not_found = " ".join(ip for ip in ips if not BlockedIP.objects.filter(ip=ip).exists())
            if not_found.strip():
                print(f"Not found: {not_found}")
                sys.exit(1)



import datetime

from django.contrib.humanize.templatetags.humanize import intcomma, naturaltime
from django.db import models
from django.db.models.fields import CharField, DateTimeField, GenericIPAddressField, IntegerField
from django.utils import timezone

DEFAULT_COOLDOWN = 7


class BlockedIP(models.Model):
    ip = GenericIPAddressField(primary_key=True, verbose_name="IP")
    first_seen = DateTimeField(default=timezone.now, db_index=True)
    last_seen = DateTimeField(blank=True, null=True, db_index=True)
    cooldown = IntegerField(
        default=DEFAULT_COOLDOWN,
        help_text="Cooldown period; number of days with no connections before IP is dropped from blocklist",
    )
    reason = CharField(blank=True, max_length=255, default="", db_index=True)
    tally = IntegerField(default=0, help_text="Number of times this IP has been blocked since first_seen")

    class Meta:
        get_latest_by = "first_seen"
        ordering = ["-last_seen", "ip"]
        verbose_name = "blocked IP"

    def __str__(self) -> str:
        return self.ip

    def verbose_str(self):
        timespan = naturaltime(self.first_seen).replace(" ago", "")
        return (
            f"{self.ip}"
            + f" -- {intcomma(self.tally)} blocks in {timespan}"
            + f" -- seen {naturaltime(self.last_seen)}"
            + f" -- {self.cooldown} day cooldown"
            + f"{' -- ' + self.reason if self.reason else ''}"
        )

    def has_expired(self):
        """Has the IP cooled long enough to be removed from the list?"""
        quiet_time = datetime.datetime.now() - self.last_seen
        return quiet_time.days >= self.cooldown

    def save(self, *args, **kwargs):
        if self.last_seen is None:
            self.last_seen = self.first_seen
        super().save(*args, **kwargs)



import datetime
import logging

from django.conf import settings
from django.http import HttpResponseBadRequest

from .models import BlockedIP
from .utils import user_ip_from_request

logger = logging.getLogger(__name__)


def denial_template():
    return (
        settings.BLOCKLIST_CONFIG.get("denial-template")
        or "Your IP address {ip} has been blocked. Try again in {cooldown} days."
    )


class BlocklistMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if entry_qs := BlockedIP.objects.filter(ip=user_ip_from_request(request)):
            entry = entry_qs.get()
            logger.warning("{} request blocked from {}".format(request.method, entry.ip))
            entry.last_seen = datetime.datetime.now()
            entry.tally += 1
            entry.save()
            return HttpResponseBadRequest(denial_template().format(ip=entry.ip, cooldown=entry.cooldown))
        else:
            return self.get_response(request)



import datetime
import logging

from django.conf import settings
from django.core.cache import cache
from django.http import HttpRequest

from .models import DEFAULT_COOLDOWN, BlockedIP

logger = logging.getLogger(__name__)

CACHE_KEY = "django-blocklist-ips"
CACHE_TTL = settings.BLOCKLIST_CONFIG.get("cache-ttl", 60)
COOLDOWN = settings.BLOCKLIST_CONFIG.get("cooldown", DEFAULT_COOLDOWN)


def user_ip_from_request(request: HttpRequest) -> str:
    """Returns user's IP. If IP can't be determined, return empty string and log a warning."""
    keys = ["HTTP_X_REAL_IP", "REMOTE_ADDR"]
    for key in keys:
        if ip := request.META.get(key):
            return ip
    # `HTTP_X_FORWARDED_FOR` is a comma-separated list with originating IP first
    if forwarded := request.META.get("HTTP_X_FORWARDED_FOR"):
        return forwarded.split(",")[0]
    logger.warning("No IP address could be found on request: {}".format(request.META))
    return ""


def get_blocklist(force_cache: bool = False) -> set:
    """Read the blocklist from cache and return it as a set. Read from database if cache is expired or `force_cache` is True."""
    blocked_ips = cache.get(CACHE_KEY)
    if blocked_ips is None or force_cache is True:
        blocked_ips = set(o.ip for o in BlockedIP.objects.all())
        cache.set(CACHE_KEY, list(blocked_ips), CACHE_TTL)
        logger.info(f"Read {len(blocked_ips)} IPs from blocklist storage; cached.")
    return set(blocked_ips)


def add_to_blocklist(ips: set, reason="") -> None:
    """Add the provided IPs to the blocklist, with optional `reason`"""
    now = datetime.datetime.now()
    for ip in ips:
        entry, new = BlockedIP.objects.get_or_create(ip=ip)  # type: ignore
        entry.reason = reason
        entry.cooldown = COOLDOWN
        entry.last_seen = now
        entry.save()
        if not new:
            logger.info(f"add_to_blocklist: updated existing entry for IP {ip}")
    logger.info(f'Added to blocklist: {ips} -- "{reason}"')
    # Update cache
    get_blocklist(force_cache=True)


def remove_from_blocklist(ip: str) -> bool:
    """Remove the IP from the blocklist. Return True if successful, False if it wasn't found."""
    try:
        BlockedIP.objects.get(ip=ip).delete()
        return True
    except BlockedIP.DoesNotExist:
        logger.warning(f"Removal of {ip} requested, but not found in blocklist.")
        return False



"""Add IPs to the blocklist from a file."""
import datetime
import logging
import sys

from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.dateparse import parse_datetime

from ...models import DEFAULT_COOLDOWN, BlockedIP

logger = logging.getLogger(__name__)
DEFAULT_DAYS = settings.BLOCKLIST_CONFIG.get("cooldown") or DEFAULT_COOLDOWN


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--dry-run", action="store_true", help="Preview the resulting update rather than performing it"
        )
        parser.add_argument(
            "--cooldown", help="'cooldown' field value for the added IPs", default=DEFAULT_DAYS
        )
        parser.add_argument("--file", help="Path to file with whitespace-separated IPs", default="ips.txt")
        parser.add_argument("--reason", help="'reason' field value for the added IPs")
        parser.add_argument(
            "--replace", action="store_true", help="Update IPs already in the DB; otherwise, skip"
        )
        parser.add_argument(
            "--timestamp",
            help="Datetime (ISO-8601) for'first_seen' and 'last_seen'; otherwise current datetime will be used",
        )

    help = __doc__

    def handle(self, *args, **options):
        for option in "dry_run cooldown file reason replace timestamp verbosity".split():
            setattr(self, option, options.get(option))
        try:
            file_handle = open(self.file)
        except FileNotFoundError as e:
            sys.exit(f"Failed to open file: {e}")
        seen = parse_datetime(self.timestamp) if self.timestamp else datetime.datetime.now()
        ips = set(file_handle.read().split())
        total = len(ips)
        self.puts(f"Read {total} unique IPs from file")
        if not self.dry_run:
            self.puts("Saving to blocklist")
        else:
            self.puts("Dry run, not saving to database")
        counts = {"added": 0, "skipped": 0}
        for n, ip in enumerate(ips):
            if len(ips) > 100 and (n * 10) % (total // 10) == 0:
                self.puts(f"Processed {n} of {total}")
            if BlockedIP.objects.filter(ip=ip).exists() and not self.replace:
                self.puts(f"Skipping {ip} - already present")
                counts["skipped"] += 1
                continue
            if not self.dry_run:
                entry = BlockedIP.objects.create(ip=ip, first_seen=seen, cooldown=self.cooldown)
                if self.reason:
                    entry.reason = self.reason
                entry.save()
                counts["added"] += 1
        self.puts(f"Done. Added {counts['added']}, skipped {counts['skipped']}")

    def puts(self, text):
        if self.verbosity > 0:
            print(text)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_blocklist',
 'django_blocklist.management',
 'django_blocklist.management.commands',
 'django_blocklist.migrations',
 'django_blocklist.tests']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2,<5']

setup_kwargs = {
    'name': 'django-blocklist',
    'version': '1.2.1',
    'description': 'A Django app that implements IP-based blocklisting.',
    'long_description': '# Django-blocklist\nThis is a [Django][] app that implements IP-based blocklisting. Its `BlocklistMiddleware` performs the blocking, and its `clean_blocklist` management command deletes entries which have satisfied the cooldown period. Entries also have a `reason` field, used in reporting. There are utility functions to add/remove IPs, an admin, and several management commands.\n\nThis app is primarily for situations where server-level blocking is not available, e.g. on platform-as-a-service hosts like PythonAnywhere or Heroku. Being an application-layer solution, it\'s not as performant as blocking via firewall or web server process, but is suitable for moderate traffic sites. It also offers better integration with the application stack, for easier management.\n\n## Quick start\n1. The [PyPI package name is `django-blocklist`](https://pypi.org/project/django-blocklist/); add that to your `requirements.txt` or otherwise install it into your project\'s Python environment.\n\n0. Add "django_blocklist" to your INSTALLED_APPS setting like this:\n\n        INSTALLED_APPS = [\n        ...\n        "django_blocklist"\n        ]\n\n0. Add the middleware like this:\n\n       MIDDLEWARE = [\n           ...\n          "django_blocklist.middleware.BlocklistMiddleware"\n       ]\n\n0. Run `python manage.py migrate` to create the `django_blocklist_blockedip` table.\n0. Add IPs to the list (via management commands,  `utils.add_to_blocklist`, or the admin).\n0. Set up a `cron` job or equivalent to run `manage.py clean_blocklist` daily.\n\n## Management commands\nDjango-blocklist includes several management commands:\n\n* `add_to_blocklist` &mdash; (one or more IPs)\n* `remove_from_blocklist` &mdash; (one or more IPs)\n* `search_blocklist` &mdash; look for an IP in the list; in addition to info on stdout, returns an exit code of 0 if successful\n* `update_blocklist` &mdash; change the `reason` or `cooldown` values for existing entries\n* `import_blocklist` &mdash; convenience command for importing IPs from a file\n* `report_blocklist` &mdash; information on the current entries\n* `clean_blocklist` &mdash; remove entries that have fulfilled their cooldown period\n\nThe `--help` for each of these details its available options.\n\nFor exporting or importing BlockedIP entries, use Django\'s built-in `dumpdata` and `loaddata` management commands.\n\n## Configuration\nYou can customize the following settings via a `BLOCKLIST_CONFIG` dict in your project settings:\n* `cooldown` &mdash; Days to expire, for new entries; default 7\n* `cache-ttl` &mdash; Seconds that utils functions cache the full list; default 60\n* `denial-template` &mdash; For the denial response; an f-string with `{ip}` and `{cooldown}` placeholders\n\n## Reporting\nThe `report_blocklist` command gives information about the current collection of IPs, including:\n* Number of listed IPs\n* Total number of blocked requests from listed IPs\n* Number of IPs active in last 24 hours\n* Number of stale IPs (added over 24h ago and not seen since)\n* Five IPs with highest block count\n* Five IPs most recently blocked\n* Longest running entry\n* IP counts by reason\n\n## Utility methods\nThe `utils` module defines two convenience functions for updating the list from your application code:\n* `add_to_blocklist(ips: set, reason="")` adds IPs to the blocklist\n* `remove_from_blocklist(ip: str)` removes an entry, returning `True` if successful\n\n[django]: https://www.djangoproject.com/',
    'author': 'Paul Bissex',
    'author_email': 'paul@bissex.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://sr.ht/~paulbissex/Django-Blocklist/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



# Generated by Django 3.2.13 on 2022-07-02 15:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_blocklist', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='blockedip',
            name='tally',
            field=models.IntegerField(default=0, help_text='Number of times this IP has been blocked since first_seen'),
        ),
    ]



"""Remove given IPs from blocklist."""
import logging

from django.core.management.base import BaseCommand

from ...models import BlockedIP

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("ips", nargs="+", type=str, help="IPs (space-separated) to remove")

    help = __doc__

    def handle(self, *args, **options):
        ips = options.get("ips")
        matches = BlockedIP.objects.filter(ip__in=ips)
        count = matches.count()
        print(f"Found {count} of {len(ips)}.")
        if count:
            matches.delete()
            print("Deleted.")



import datetime

from django.core.cache import cache
from django.test import TestCase

from ..middleware import denial_template
from ..models import BlockedIP
from ..utils import CACHE_KEY


class MiddlewareTests(TestCase):
    def setUp(self):
        BlockedIP.objects.all().delete()
        cache.delete(CACHE_KEY)

    def tearDown(self):
        BlockedIP.objects.all().delete()
        cache.delete(CACHE_KEY)

    def test_ip_block_hit(self):
        BlockedIP.objects.create(ip="127.0.0.1")
        response = self.client.get("/")
        self.assertEqual(response.status_code, 400)

    def test_ip_block_message(self):
        entry = BlockedIP.objects.create(ip="127.0.0.1")
        response = self.client.get("/")
        expected = denial_template().format(ip=entry.ip, cooldown=entry.cooldown)
        self.assertTrue(expected in str(response.content))

    def test_ip_block_miss(self):
        response = self.client.get("/")
        self.assertEqual(response.status_code, 200)

    def test_update_last_seen(self):
        then = datetime.datetime(2020, 2, 29)
        BlockedIP.objects.create(ip="127.0.0.1", first_seen=then, last_seen=then, reason="last-seen test")
        self.client.get("/")
        self.assertGreater(BlockedIP.objects.get(reason="last-seen test").last_seen, then)

    def test_update_tally(self):
        BlockedIP.objects.create(ip="127.0.0.1")
        self.client.get("/")
        self.client.get("/")
        entry = BlockedIP.objects.get(ip="127.0.0.1")
        self.assertEqual(entry.tally, 2)



"""Print summary information about the data in the blocklist."""
import datetime
from collections import Counter
from operator import itemgetter
from typing import Iterable, Tuple

from django.contrib.humanize.templatetags.humanize import intcomma
from django.core.management.base import BaseCommand, CommandError
from django.db.models import Sum

from ...models import BlockedIP


class Command(BaseCommand):
    help = __doc__

    def handle(self, *args, **options):
        entries = BlockedIP.objects.all()
        if entries.count() == 0:
            raise CommandError("No BlockedIP objects in database.")
        _grand_tally = BlockedIP.objects.all().aggregate(Sum("tally"))["tally__sum"]
        print(f"Total blocks of listed IPs: {intcomma(_grand_tally)}")
        print(f"Entries in blocklist: {intcomma(entries.count())}")
        _one_day_ago = datetime.datetime.now() - datetime.timedelta(days=1)
        print(f"Active in last 24 hours: {intcomma(entries.filter(last_seen__gte=_one_day_ago).count())}")
        print(
            f"Stale (added over 24h ago, not seen since): {intcomma(entries.filter(tally=1, last_seen__lt=_one_day_ago).count())}"
        )
        print()
        print_roster("Most active", entries.exclude(tally=0).order_by("-tally")[:5])
        print_roster("Most recent", entries.exclude(tally=1).order_by("-last_seen")[:5])
        longest_lived = None
        how_long = datetime.timedelta(0)
        for entry in BlockedIP.objects.all():
            active_period = entry.last_seen - entry.first_seen
            if active_period > how_long:
                longest_lived, how_long = entry, active_period
        if longest_lived is not None:
            print(f"Longest lived:\n{longest_lived.verbose_str()}")
        reasons = reason_counts()
        if reasons:
            print("\nIP counts by reason\n-------------------")
            _width = len(str(reasons[0][1])) + 1
            for reason, count in reasons:
                print(f"{count: {_width}} | {reason}")


def print_roster(title: str, queryset):
    print(f"{title}:")
    for perp in queryset:
        print(perp.verbose_str())
    print()


def reason_counts() -> Iterable[Tuple[str, int]]:
    reason_data = Counter(BlockedIP.objects.exclude(reason="").values_list("reason"))
    tuples = [(str(r[0][0]), r[1]) for r in reason_data.items()]
    return sorted(tuples, key=itemgetter(1), reverse=True)
